 package alignment;
 
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import datalayer.OboHomologyWriter;
import datalayer.WorkspaceWriter;

import ontology.Annotations;
import ontology.HomologyOntology;
import ontology.Ontology;
import ontology.OntologyInstance;
import ontology.SpeciesOntology;
import visualization.Main;

// /!\ TODO LIST:
//
// - Dans la prise en compte des groupes d'homologies, il y a un cas particulier que je ne
//   regarde pas : celui ou un groupe d'homologie est trouve par homonymie, qui contient
//   un organe de l'espece de reference, mais que celui-ci n'est pas un homonyme de
//   l'organe de travail. Dans ce cas precis le groupe d'homologie est pondere comme s'il
//   ne contenait pas d'organe de l'espece de reference et je ne suis pas certain que ce
//   soit la meilleure solution.

/**
 * HomologyMaker is the class for the main alignment algorithm. This algorithm is based on
 * homonymy search, with propagation of homonymy score in the hierarchy. The idea is that
 * if two terms are homonyms, then they probably are homologous and thus there parents
 * could be homologous too.
 *
 * @author Gilles Parmentier
 */
public class HomologyMaker extends IterativeAligner
{
 // Ontologies:
 SpeciesOntology _oReference, _oWorking;
 HomologyOntology _oHomology;

 // Data structures:
 LinkedList <AnnotatedOrgan> _annotatedOrgans;
 PairingScorer _pairing_scorer;
 Vector <AnnotatedOrgan> _propositions;
 Vector <Annotations> _annotations ;
 Vector <PairingScore> _discards;
 HomologyValidator _validator;

 // Data structures for user input done prior to the algorithm:
 Vector <Vector <String> > _homology_created;
 Vector <Boolean> _strict_flags;
 Vector <Vector <String> > _invalidate_created;
 Main _m;

 private static final long serialVersionUID = 1L;
 
 public HomologyMaker (Main m)
 {
  _m = m;
  _propositions = null;

  _annotations = new Vector <Annotations> ();
  _homology_created = new Vector <Vector <String>> ();
  _invalidate_created = new Vector <Vector <String>> ();
  _strict_flags = new Vector <Boolean> ();
  _discards = new Vector <PairingScore> ();
  }

 /**
  * Initialises the data structures for the alignment process.
  * @return A <code>Vector</code> of all the <code>AnnotatedOrgan</code> that already are
  * member of an homology group that contains at least one organ of the reference ontology.
  */
 private Vector <AnnotatedOrgan> init (Vector <Vector <String>> homologies_added, Vector <Boolean> flags)
 {
  System.err.println (this.getClass ().getName () + "::init (): ");
  _annotatedOrgans = new LinkedList<AnnotatedOrgan> ();
  Vector <AnnotatedOrgan> res = new Vector<AnnotatedOrgan> ();
  LinkedList <OntologyInstance> workingInstances = _oWorking.getInstancesAsLinkedList ();

  // Ici il faut que je verifie simplement si l'organe appartient deja a un groupe
  // d'homologie. L'autre recherche devrait etre deplacee a la partie en dessous, elle fait
  // partie de la premiere etape de l'alignement.
  // La partie ici est une initialisation.

  // - First, create the AnnotatedOrgan for all the organs of the working ontology,  and
  //   check if the corresponding organ is member of an existing homology group.
  String partof_id = _oHomology.getPartOfPropertyId ();

  for (int i = 0; i < workingInstances.size (); i++)
      {
       AnnotatedOrgan ao = new AnnotatedOrgan (_m, this, _oHomology,
                                             (OntologyInstance) workingInstances.get (i));
       addToAnnotatedOrgans (ao);

       // Ici je dois verifier si l'organe appartient deja a un groupe d'homologie.
       List <OntologyInstance> hg = _oHomology.getInstances();
       for (int hg_index = 0; hg_index < hg.size (); hg_index++)
           {
            OntologyInstance group = hg.get (hg_index);
            //System.out.println ("Verifying " + group.getId ());
            if (group.hasInstanceAsSubjectOfProperty ((OntologyInstance) workingInstances.get (i), "inv_" + partof_id))
               {
                System.out.println (((OntologyInstance) workingInstances.get (i)).getId () + " is in an homology group (" + group.getId () + ")");
                Vector <String> s_organs = group.getObjectOfPropertyByPropertyIdAsVector ("inv_" + partof_id);
                //System.out.println ("It contains " + s_organs.size () + " organs");
                boolean added = false;
                for (int org_index = 0; org_index < s_organs.size (); org_index++)
                    {
                     // Ajouter les organes de l'ontologie de reference comme homologues.
                     // Les mettre dans la liste d'exception. Prevoir de propager leurs
                     // scores.
                     OntologyInstance oi = _oReference.getInstanceById ((String) s_organs.get(org_index));
                     if (oi != null)
                        {
                         //System.out.println ("found an homolog");
                         ao.setHomolog (group, oi);
                         added = true;
                         }
                     else
                        {
                    	 //System.out.println ("No instance found for " +
                         //                               (String) s_organs.get(org_index));
                    	 }
                     }
                if (added)
                   res.add (ao);
                }
            }
       }

  // Maintenant je dois m'occuper des homologies qui ont ete ajoutees manuellement par
  // l'utilisateur avant de commencer l'alignement. Elles ne peuvent pas etre traitees
  // comme des homologies existantes car elles peuvent provoquer la fusion de groupes.
  for (int i = 0; i < homologies_added.size (); i++)
      {
       //System.err.println ("Adding an homology");
       String working_id = homologies_added.get (i).get (0);
       OntologyInstance woi = _oWorking.getInstanceById (working_id);
       String reference_id = homologies_added.get (i).get (1);
       AnnotatedOrgan wao = this.getAnnotatedOrganByOrganId (working_id);
       boolean flag = flags.get (i);
 
       // Je tente en ne creant qu'un PairingScore fictif et en validant le tout.
       PairingScore ps = new PairingScore (woi);
       ps.setCandidate (_oReference.getInstanceById (reference_id), -1);
       wao.addPairingScore (ps);
       wao.validateProposition (ps, flag);
       }

  //this.dumpAnnotatedOrgans ();
  System.err.println ("End of HomologyMaker::init");
  return res;
  }
 
 public void align (PairingScorer ps)
 {
  if (ps == null)
     System.err.println (this.getClass ().getName () + "::align (): ps is null!");
  _pairing_scorer = ps;
  //System.out.println ("init");
  align (_homology_created, _strict_flags, _invalidate_created);
  }

 /**
  * 
  * @param discards
  * @param _homology_created
  * @param flags
  * @param _invalidate_created
  * @return
  */
 public boolean align (Vector <Vector <String>> _homology_created, Vector <Boolean> flags, Vector<Vector<String>> _invalidate_created)
 {
  _pairing_scorer.init (new Vector <OntologyInstance> (_oWorking.getInstances ()),
                             new Vector <OntologyInstance> (_oReference.getInstances ()));

 init (_homology_created, flags);

  // - Second, for each organ, look for homonyms in their reference ontology.

  // Verifier si on a des homonymes dans l'ontologie de reference :
  // Attention, ici il ne faut pas compter les homonymes appartenant deja a un groupe
  // d'homologie ! Ou alors je les compte deux fois, ce qui augmente artificiellement le
  // score des homonymes deja valide dans un groupe d'homologie. Intuitivement j'ai envie
  // de donner un plus gros score a ces derniers.
  int organ_with_homonym = 0;
  for (int i = 0; i < _annotatedOrgans.size (); i++)
      {
       Vector <OntologyInstance> homonyms = _oReference.getInstancesByNameAsVector (
                                                     _annotatedOrgans.get (i).getName ());

       if (homonyms.size () > 0)
          {
           //System.out.println ("We found " + homonyms.size () + " homonyms for " +
           //                       _annotatedOrgans.get (i).getName ());

    	   //System.out.println ("Organ: " + _annotatedOrgans.get (i).getOrganId () + " (" + _annotatedOrgans.get (i).getName () + ") " + homonyms);

           for (int j = 0; j < homonyms.size (); j++)
               _annotatedOrgans.get (i).addHomonym (
                                               (OntologyInstance) homonyms.elementAt (j));
           organ_with_homonym++;
           }

       // Check if this organ has homonyms in some homology groups:
       // At this point we only have to search in the provided HomologyOntology.
       Vector <OntologyInstance> hgs = _oHomology.getHomologyGroupsByName (_annotatedOrgans.get (i).getOrgan ().getName ());
       if (hgs.size () > 0)
          {
           System.out.println ("HomologyMaker::align: Found homology group,nothing done");
           // /!\ TODO: Pour chaque groupe d'homologie, creer un nouveau score pour notre
           //           AnnotatedOrgan. Si un des organes de l'espece de reference
           //           appartient a ce groupe.
           }
       }
  System.out.println (organ_with_homonym + " organs have at least one homonym.");

  // Now we should look for homonyms in the synonyms lists:
  // Three categories:
  // - names of the reference ontology with synonyms of the working ontology,
  // - synonyms of the reference ontology with the names of the working ontology,
  // - synonyms of the reference ontology with synonyms of the working ontology.
  //System.err.println ("Checking synonyms");
  for (int i = 0; i < _annotatedOrgans.size (); i++)
      {
       Vector <OntologyInstance> homonyms = _oReference.getInstancesBySynonymAsVector (_annotatedOrgans.get (i).getName ());

       // First case: instance 1, name; instance 2, synonym
       if (homonyms.size () > 0)
          {
	   //System.out.println (_annotatedOrgans.get (i).getOrganId() + " has name-synonym");
           for (int j = 0; j < homonyms.size (); j++)
               _annotatedOrgans.get (i).addSynonym (
                                          (OntologyInstance) homonyms.elementAt (j), null,
                                                     _annotatedOrgans.get (i).getName ());
           organ_with_homonym++;
           }
       // Second case: we check synonyms of the first instance, two possibilities: match with the name or a synonym of the 2nd instance.
       Vector <String> synonyms = _annotatedOrgans.get (i).getOrgan ().getObjectOfPropertyByPropertyIdAsVector ("synonym");
       if (synonyms != null)
          {
           for (int j = 0; j < synonyms.size (); j++)
               {
                // Check with names:
                Vector <OntologyInstance> homonyms2 = _oReference.getInstancesByNameAsVector (synonyms.get (j));
                for (int k = 0; k < homonyms2.size (); k++)
                    {
         	   //System.out.println (_annotatedOrgans.get (i).getOrganId() + " has synonym-name");
                    _annotatedOrgans.get (i).addSynonym (
                      (OntologyInstance) homonyms2.elementAt (k), synonyms.get (j), null);
                organ_with_homonym++;
                    }
                // Check with synonyms:
                homonyms2 = _oReference.getInstancesBySynonymAsVector (synonyms.get (j));
                for (int k = 0; k < homonyms2.size (); k++)
                {
          	 //System.out.println (_annotatedOrgans.get (i).getOrganId() + " has synonym-synonym; " + synonyms.get (j));
                 _annotatedOrgans.get (i).addSynonym ((OntologyInstance) homonyms2.elementAt (k), synonyms.get (j), synonyms.get (j));
                organ_with_homonym++;
                }
                }
           }
       }

  // Propagation of the scores through contigous homonymous pairs
  _validator = new HomologyValidator (this, _annotatedOrgans, _oReference.getInstancesAsLinkedList (), _pairing_scorer);
  _validator.propagateHomonymy ();

  cleanCandidateList ();

  dumpAnnotatedOrgans ();

  System.out.println ("--- End of HomologyMaker::align");

  return true;
  } // public boolean align ()

 /**
  * 
  *
  */
 private void cleanCandidateList ()
 {
  double epsilon = 0.001;
  System.out.println ("Cleaning candidates lists");
  for (int i = 0; i < _annotatedOrgans.size (); i++)
      {
       AnnotatedOrgan ao = _annotatedOrgans.get (i);
       Vector <PairingScore> vps = ao.getScores ();
       //System.out.println ("Organ is: " + ao.getName () + " (" + ao.getOrganId () + "): " + vps.size () + " candidates");
       if (vps.size () > 1)
          {
           double sum = 0;
           double max = 0;
           int nb_max = 0;
           for (int j = 0; j < vps.size (); j++)
               {
                PairingScore ps = vps.get (j);
                sum += ps.getScore ();
                //ystem.out.println (ps.getScore () + "; max: " + max + "; " + nb_max);
                if (ps.getScore () > max + epsilon)
                   {
                    max = ps.getScore ();
                    nb_max = 1;
                    }
                else
                   if ((ps.getScore () > max - epsilon) && ( ps.getScore () < max + epsilon))
                      {
                       max = ps.getScore ();
                       nb_max++;
                       //System.err.println ("New max");
                       }
                }
           //System.out.println ("Average: " + sum / vps.size () + "; max: " + max + "; nb_max: " + nb_max);
           Vector <PairingScore> new_vps = new Vector <PairingScore> ();
           if (nb_max != vps.size ())
              {
               for (int j = 0; j < vps.size (); j++)
                   {
                    PairingScore ps = vps.get (j);
                    if (ps.getScore () > max - epsilon)
                       {
                        new_vps.add (vps .get (j));
        	            }
                    }
               ao.setScores (new_vps);
               //System.out.println ("final size: " + new_vps.size ());
               }
           else
              {
               // All candidates have the same score. If we have too many of them, juste
               // remove them.
               if (nb_max > 5)
                  {
                   ao.setScores (new Vector <PairingScore> ());
                   }
               }
           }
       }
 }

 /**
  * Adds an AnnotatedOrgan to the list of AnnotatedOrgan.
  *
  * @param ao The AnnotatedOrgan to add to this object.
  */
 // Wrapper in the case I would like to have a specific insertion order.
 void addToAnnotatedOrgans (AnnotatedOrgan ao) { _annotatedOrgans.add (ao); }

 public void createInvalidate (String working_id, String reference_id)
 {
  AnnotatedOrgan ao = this.getAnnotatedOrganByOrganId (working_id);
  if (ao != null)
     {
      ao.invalidatePossibility (reference_id);
      }
//  if (_aligner == null)
//  {
//   Vector <String> v = new Vector<String> ();
//   v.add (working_id);
//   v.add (reference_id);
//   _invalidate_created.add (v);
//   }
//else
//  {
//   // L'alignement a deja ete commence par l'utilisateur, il faut donc que "j'injecte" directement cette info dans l'alignement en cours.
//	  _aligner.createInvalidate (working_id, reference_id);
//   }
  }
 
 /**
  * Not used
  */
 public boolean iterate (Vector <PairingScore> v) { return false; }

 /**
  *
  */
 public boolean iterate (Vector <PairingScore> validate_strict, Vector <PairingScore> validated, Vector <PairingScore> invalidated, Vector <PairingScore> delay_ignore)
 {
  System.err.println ("Iteration: " + validate_strict.size () + " strict validations; " + validated.size () + " validations; " + invalidated.size () + " invalidations");
  for (int i = 0; i < invalidated.size (); i++)
      {
       _discards.add ((PairingScore) invalidated.get (i));
       _validator.invalidate ((PairingScore) invalidated.get (i));
       }

  for (int i = 0; i < delay_ignore.size (); i++)
      {
       _validator.delay ((PairingScore) delay_ignore.get (i));
       }

  for (int i = 0; i < validate_strict.size (); i++)
      {
       _validator.validate ((PairingScore) validate_strict.get (i), true);
       }
  for (int i = 0; i < validated.size (); i++)
      {
       _validator.validate ((PairingScore) validated.get (i), false);
       }
  //this.dumpAnnotatedOrgans();
  return false;
  }

 /**
  * Returns an ontology instance of the reference ontology given its id.
  * @param ref_id The id of the wanted OntologyInstance
  * @return The OntologyInstance, <code>null</code> if not found.
  * @see OntologyInstance
  */
 OntologyInstance getReferenceOntologyInstanceById (String ref_id)
 {
  return  _oReference.getInstanceById (ref_id);
  }

 /**
  * Returns the AnnotatedOrgan for the organ for wich the id is provided.
  * @param o_id The reference of the organ.
  * @return The <code>AnnotatedOrgan</code> corresponding to the given Id,
  * <code>null</code> if the corresponding organ does not have an associated
  * <code>AnnotatedOrgan</code> yet.
  */
 public AnnotatedOrgan getAnnotatedOrganByOrganId (String o_id)
 {
  for (int i = 0; i < _annotatedOrgans.size (); i++)
      if (_annotatedOrgans.get (i).getOrganId ().equals (o_id))
         return _annotatedOrgans.get (i);
  return null;
  }

 /**
  * Debugging function. Dump the annotated organs in the console in order to be able to
  * check the results after an iteration.
  */
 void dumpAnnotatedOrgans ()
 {
  Vector<AnnotatedOrgan> propositions = new Vector<AnnotatedOrgan> (_annotatedOrgans);
  Collections.sort (propositions, new Comparator<AnnotatedOrgan> ()
                                  { public int compare (AnnotatedOrgan ao1,
                                                                       AnnotatedOrgan ao2)
                                           { return (int) ((ao2.getBestScore () -
                                                        ao1.getBestScore ()) * 100); } });

  int number_of_organs_with_candidate = 0;
  for (int i = 0; i < propositions.size (); i++)
      if (propositions.get (i).dumpToConsole ())
         number_of_organs_with_candidate++;
  System.out.println ("Number of organs with at least one candidate homolog: " +
                                                         number_of_organs_with_candidate);
  }

 /**
  * Returns the best results of the alignement. These proposed appariements should be
  * validated by the user. This function should accept a parameter, the number of
  * propositions we want to make to the user.
  * @return A <code>Vector</code> of <code>AnnotatedOrgan</code>, the ones with the
  * highests scores.
  */
 public Vector<AnnotatedOrgan> getPropositions ()
 {
  _propositions = new Vector<AnnotatedOrgan> ();
  for (int i = 0; i < _annotatedOrgans.size (); i++)
      {
       AnnotatedOrgan ao = _annotatedOrgans.get (i);
       _propositions.add (ao);
       }
  Collections.sort (_propositions, new Comparator<AnnotatedOrgan> ()
                             { public int compare (AnnotatedOrgan ao1, AnnotatedOrgan ao2)
                 { return (int) ((ao2.getBestScore () - ao1.getBestScore ()) * 100); } });

  return _propositions;
  }



 /**
  *
  **/
 public void saveDiscards (String filename) throws IOException
 {
  // Discards:
  try {
       FileWriter file = new FileWriter (new File (filename));
       BufferedWriter bw = new BufferedWriter (file);
       LinkedList<AnnotatedOrgan> aos = getAnnotatedOrgans ();
       if (aos == null)
          {
           bw.close ();
           return;
           }
       for (int i = 0; i < aos.size (); i++)
           {
            AnnotatedOrgan ao = aos.get (i);
            ao.writeDiscards (bw);
            }
       bw.close ();
       }
     catch (Exception e)
           { System.out.println ("failure"); e.printStackTrace(); }
 // Scoring function:
 try {
      // Premierement je dois voir si le fichier existe. Si c'est le cas je recupere les
      // anciennes valeurs. J'ajoute les nouvelles valeurs et je sauve le tout.
      Vector<String[]> values = new Vector<String[]> ();
      File f = new File (filename + "_values.oah");
      if (f.exists ())
         {
          FileReader fr = new FileReader (f);
          BufferedReader br = new BufferedReader (fr);
  String str = br.readLine ();
  while (str != null)
        {
         String [] words = str.split ("\t");
         values.add (words);
         str = br.readLine ();
         }
  br.close ();

  BufferedWriter bw = new BufferedWriter (new FileWriter (new File (filename + "_values.oah")));
  LinkedList<AnnotatedOrgan> aos = getAnnotatedOrgans ();
  for (int i = 0; i < aos.size (); i++)
      {
       AnnotatedOrgan ao = aos.get (i);
       ao.writeValues (bw, values);
       }
  bw.close ();
          }
      else
         {
          // Ici j'ai juste a ecrire les nouvelles valeurs.
          BufferedWriter bw = new BufferedWriter (new FileWriter (new File (filename + "_values.oah")));
          LinkedList<AnnotatedOrgan> aos = getAnnotatedOrgans ();
          for (int i = 0; i < aos.size (); i++)
              {
               AnnotatedOrgan ao = aos.get (i);
               ao.writeValues (bw, null);
               }
          bw.close ();
          }
      }
 catch (Exception e)
 { System.out.println ("failure"); e.printStackTrace(); }
 }

 /**
  * 
  * @param filename
  */
 public void saveHomologyOntology (String filename)
 {
  _oHomology.setBaseURI (filename);
  OboHomologyWriter ohw = new OboHomologyWriter (filename, _oHomology);
  }

 public void saveWorkspace (String f, String disc, String sannot)
 {
  //WorkspaceWriter ww = new WorkspaceWriter (f, disc, this._oHomology, sannot);
  WorkspaceWriter ww = new WorkspaceWriter (f);
  ww.write (this._oHomology);
  }
 
 public void saveWorkspace(String _workspace_filename, String _homology_ontology_filename, String _discards_filename, String sannot)
 {
  WorkspaceWriter ww = new WorkspaceWriter (_workspace_filename, _homology_ontology_filename, _discards_filename, this._oHomology, sannot);
  }
 
 // Getters:
 
 public LinkedList<AnnotatedOrgan> getAnnotatedOrgans () { return _annotatedOrgans; }

 public PairingScorer getPairingScorer () { return _pairing_scorer; }

 public Vector <Annotations> getAnnotations () { return _annotations; }

 public Ontology getReferenceOntology () { return _oReference; }

 public Ontology getWorkingOntology () { return _oWorking; }

 public Ontology getHomologyOntology () { return _oHomology; }

 // Setters:

 public void setScorer (PairingScorer ps) { this._pairing_scorer = ps; }
 
 public void setReferenceOntology (SpeciesOntology r) { this._oReference = r;
 if (_oHomology != null) _oHomology.AddSpeciesURI (_oReference.getBaseURI ());
  }

 public void setWorkingOntology (SpeciesOntology w) { this._oWorking = w;
 if (_oHomology != null) _oHomology.AddSpeciesURI (_oWorking.getBaseURI ()); }

 /**
  * 
  * @param h
  */
 public void setHomologyOntology (HomologyOntology h)
 {
  this._oHomology = h;
  if (_oReference != null) _oHomology.AddSpeciesURI (_oReference.getBaseURI ());
  if (_oWorking != null) _oHomology.AddSpeciesURI (_oWorking.getBaseURI ());
  }

 // Testers:

 public boolean hasDiscards () { return (_discards.size () > 0); }

 public void setAnnotations (Vector <Annotations> annot) { this._annotations = annot; }

@Override

 public boolean align () { return false; }

 // TODO:
 public void createHomology (String working_id, String reference_id, boolean strict)
 {
  // Here I assume that the alignement didnt start, to be changed:
  Vector <String> v = new Vector <String> ();
  v.add (working_id);
  v.add (reference_id);
  _homology_created.add (v);
  _strict_flags.add (strict);
  }

 /**
  * 
  * @param reference
  * @param worker
  * @param author
  * @param link
  * @param comments
  * @param i
  */
 public void createAnnotation (String reference, String worker, String author, String link, String comments, int i)
 {
  Annotations a = new Annotations (_oReference.getInstanceById (reference), _oWorking.getInstanceById (worker), author, link, comments, i);
  _annotations.add (a);
  }
 }