package alignment;

import java.util.LinkedList;
import java.util.Vector;

import ontology.OntologyInstance;
import ontology.SpeciesOntology;

public class HomologyValidator
{
 SimpleHomologyPropagator _propagator;
 LinkedList <AnnotatedOrgan> _annotated_organs;
 LinkedList <OntologyInstance> _reference_organs;
 PairingScorer _scorer;
 HomologyMaker _maker;

 public HomologyValidator (HomologyMaker m, LinkedList <AnnotatedOrgan> annotated_organs, LinkedList <OntologyInstance> reference_organs, PairingScorer scorer)
 {
  _maker = m;
  _propagator = new SimpleHomologyPropagator (_maker);
  _scorer = scorer;
  _annotated_organs = annotated_organs;
  _reference_organs = reference_organs;
  }

 /**
  * Validates an homology. This consists in two steps: first, create an homology group
  * containing the two instances that are validated. Second, propagate the results of the
  *  validation in the hierarchy.
  *  @param ps
  *  @param strict
  */
 public void validate (PairingScore ps, boolean strict)
 {
  AnnotatedOrgan ao = null;
  int pos = 0;
  for (int i = 0; i < _annotated_organs.size (); i++)
      {
       ao = (AnnotatedOrgan) _annotated_organs.get (i);
       if (ao.getOrgan ().getId ().equals (ps.getInstance ().getId ()))
          {
           pos = i;
           break;
           }
       }

  _propagator.removePropagatedScore (ao, pos, ps);
  ao.validateProposition (ps, strict);
  OntologyInstance candidate = ps.getOrganCandidate ();

  _propagator.propagateForOrgan (ao, ps, candidate);
  }

 /**
  *
  */
 public void invalidate (PairingScore ps)
 {
  AnnotatedOrgan ao = null;
  int pos = 0;
  for (int i = 0; i < _annotated_organs.size (); i++)
      {
       ao = (AnnotatedOrgan) _annotated_organs.get (i);
       if (ao.getOrgan ().getId ().equals (ps.getInstance ().getId ()))
          {
           pos = i;
           break;
           }
       }
  String old_name = ps.getOrganCandidate ().getName ();
  _propagator.removePropagatedScore (ao, pos, ps);
  // Decrease the score of the invalidated pair:
  _scorer.decreasePairingScore (ao.getOrgan ().getName (), old_name);
  PairingScore new_ps = ao.invalidateProposition (ps);
  if (new_ps != null)
     {
      OntologyInstance candidate = new_ps.getOrganCandidate ();
      _propagator.propagateForOrgan (ao, ps, candidate);
      }
  }

 /**
  * 
  * @param ps
  */
 public void delay (PairingScore ps)
 {
  AnnotatedOrgan ao = null;
  int pos = -1;
  for (int i = 0; i < _annotated_organs.size (); i++)
      {
       ao = (AnnotatedOrgan) _annotated_organs.get (i);
       if (ao.getOrgan ().getId ().equals (ps.getInstance ().getId ()))
          {
           pos = i;
           break;
           }
       }

  _propagator.removePropagatedScore (ao, pos, ps);
  PairingScore new_ps = ao.delayProposition (ps);
  if (new_ps != null)
     {
      OntologyInstance candidate = new_ps.getOrganCandidate ();
      _propagator.propagateForOrgan (ao, ps, candidate);
      }
  }
 
 /**
  * 
  */
 public void propagateHomonymy ()
 {
 System.out.println ("HomologyValidator::propagateHomonymy");
  // Propagation only. This is for the initialization step, we propagate the scores of the
  // homonymy. Currently we do it without creating new pairings.
  // For all organs of the left ontology:
  for (int i = 0; i < _annotated_organs.size (); i++)
      {
       // Check if the current organ has pairings:
       AnnotatedOrgan ao = (AnnotatedOrgan) _annotated_organs.get (i);
       if (ao.hasCandidate ())
          {
           // ao has at least one pairing.
           // Check if any child as a pairing with a child of the right organ. If so, we propagate the score of the pairing to this children pairing.
           PairingScore ps_current = ao.getBestScoreObject ();
           OntologyInstance right = ps_current.getOrganCandidate ();
           OntologyInstance left = ao.getOrgan ();
           System.out.println ("Checking " + left.getName () + " (" + left.getId () + ") " + right.getName () + "(" + right.getId () + ")");
           Vector <String> children_id = left.getObjectOfPropertyByPropertyIdAsVector (
                 "inv_" + ((SpeciesOntology) left.getOntology ()).getPartOfPropertyId ());
           // Check for each child of the left organ, check if he has at least one pairing:
           if ((children_id != null) && (children_id.size () > 0))
              for (int j = 0; j < children_id.size (); j++)
               {
                AnnotatedOrgan ao_child =
                                  _maker.getAnnotatedOrganByOrganId (children_id.get (j));
                if (ao_child.hasCandidate ())
                   {
                    System.out.println ("\t" + (j + 1) + "/" + children_id.size () + "(" + ao_child.getOrgan ().getId () + "; " + ao_child.getOrgan ().getName () + ")" + "; has candidate");
                    // Check if one of the candidate is a son of right:
                    Vector <String> right_children_id = right.getObjectOfPropertyByPropertyIdAsVector ("inv_" + ((SpeciesOntology) right.getOntology ()).getPartOfPropertyId ());
                    Vector <PairingScore> ps = ao_child.getScores ();
                    if (right_children_id != null)
                    System.out.println ("\tright has " + right_children_id.size () + " sons");
                    for (int k = 0; k < ps.size (); k++)
                        {
                	     OntologyInstance child_candidate = ps.get (k).getOrganCandidate ();
                        System.out.println ("\tcandidate " + (k + 1) + "/" + ps.size () + "(" + child_candidate.getId () + "; " + child_candidate.getName () + ")");
                 // Check if child_candidate is in right_children_id:
              	   if (right_children_id != null)
                   {
                    int l = 0;
                	 for (; l < right_children_id.size (); l++)
                	     {
                	      if (child_candidate.getId ().equals (right_children_id.get (l)))
                	         {
                           System.out.println ("\t=> it is a son of right, increasing score; added " + ps_current.getScore ());
                		  // Propagate the score of ps_current to ps:
                		  ps.get (k).addToReceivedScore (ps_current.getScore (), ps_current.getInstance ().getId () + ps_current.getOrganCandidate ().getId ());
                         System.out.println ("\t=> New score is " + ps.get (k).getScore ());
                          break;
                	          }
                	      }
                     //if (l >= right_children_id.size ()) { System.err.println ("\t=> Not a son of right."); }
                   }
                         }
                    }
                }
       //else
	      //System.err.println ("\tNo child");

           Vector <String> parents_id = left.getObjectOfPropertyByPropertyIdAsVector (
             ((SpeciesOntology) left.getOntology ()).getPartOfPropertyId ());
       if ((parents_id != null) && (parents_id.size () > 0))
          for (int j = 0; j < parents_id.size (); j++)
           {
            AnnotatedOrgan ao_parent =
                              _maker.getAnnotatedOrganByOrganId (parents_id.get (j));
            if (ao_parent.hasCandidate ())
               {
                Vector <String> right_parent_id = right.getObjectOfPropertyByPropertyIdAsVector (((SpeciesOntology) right.getOntology ()).getPartOfPropertyId ());
                Vector <PairingScore> ps = ao_parent.getScores ();
                for (int k = 0; k < ps.size (); k++)
                    {
                     OntologyInstance parent_candidate = ps.get (k).getOrganCandidate ();
               if (right_parent_id != null)
               {
                int l = 0;
                 for (; l < right_parent_id.size (); l++)
                     {
                      if (parent_candidate.getId ().equals (right_parent_id.get (l)))
                         {
                      ps.get (k).addToReceivedScore (ps_current.getScore (), ps_current.getInstance ().getId () + ps_current.getOrganCandidate ().getId ());
                      break;
                          }
                      }
               }
                     }
                }
            }
           }
       }
  System.out.println ("!!!! End ofHomologyValidator::propagateHomonymy");
  }
 }
