package alignment;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;

import ontology.OntologyInstance;

// Le calcul du score se decompose en deux parties.
// La premiere partie est intrinseque a un organe : elle consiste en l'homonymie, avec un
// groupe d'homologie et/ou un organe. Plus tard elle incluera egalement une partie due a
// la similitude dans le voisinage avec un organe. Cette partie est ce qui est propage.
// La deuxieme partie du score est due a la propagation. Cette partie n'est pas propagee.

/**
 * How to keep trace of a putative homology grouping. Currently this is not design to
 * store a confirmed homology, but I think it could.
 *
 * @author Gilles Parmentier
 */
public class PairingScore implements Serializable
{
 /**
  *
  */
 private static final long serialVersionUID = 1L;

 /**
  * Stores the intrinsic score of this putative coupling. It is based on homonymy and
  * eventually homology group.
  * It is a dynamic value : it depends on the choices made by the expert. So this value is
  * only used for ad-hoc homology groups, coming from a preceding session or specified by
  * the user.
  */
 double _score;
 double _receivedScore;
 OntologyInstance _homologyGroup;
 OntologyInstance _candidate;
 boolean _isValidated;
 OntologyInstance _instance;
 boolean _hasReceived = false;
 boolean _initialCandidate = true;
 boolean _initialSynonym1Candidate = true;
 boolean _initialSynonym2Candidate = true;
 Hashtable <String, Double> _receivedScoreContribution;
 String _synonym1;
 String _synonym2;
 
 Vector <PairingScore> _group;

 static PairingScorer _scorer;

 PairingScore (OntologyInstance oi)
 {
  _score = -1;
  _homologyGroup = null;
  _candidate = null;
  _receivedScore = - 1;
  _isValidated = false;
  _instance = oi;
  _receivedScoreContribution = new Hashtable <String, Double> ();
  _initialSynonym1Candidate = false;
  _initialSynonym2Candidate = false;
  }

 public PairingScore (OntologyInstance left, OntologyInstance right, double score)
 {
  _instance = left;
  _candidate = right;
  _score = score;
  _homologyGroup = null;
  _receivedScore = - 1;
  _isValidated = true;
  _initialCandidate = false;
  _initialSynonym1Candidate = false;
  _initialSynonym2Candidate = false;
  }

 void setCandidate (OntologyInstance c)
 {
  _candidate = c;
  //System.err.println ("PairingScore::setCandidate (): " + c.getId () + " for " + _instance.getId ());
  _initialCandidate = false;
  _initialSynonym1Candidate = false;
  _initialSynonym2Candidate = false;
  }

 void setInitialCandidate (OntologyInstance c)
 {
  _candidate = c;
  _initialCandidate = true;
  _initialSynonym1Candidate = false;
  _initialSynonym2Candidate = false;
  }

 void setInitialSynonym1Candidate (OntologyInstance c, String s1)
 {
  _candidate = c;
  _initialCandidate = true;
  _initialSynonym1Candidate = true;
  _initialSynonym2Candidate = false;
  _synonym1 = s1;
  }

 void setInitialSynonym2Candidate (OntologyInstance c, String s2)
 {
  _candidate = c;
  _initialCandidate = true;
  _initialSynonym1Candidate = false;
  _initialSynonym2Candidate = true;
  _synonym2 = s2;
  }

 void setInitialSynonym12Candidate (OntologyInstance c, String s1, String s2)
 {
  //System.out.println ("setInitialSynonym12Candidate " + s1 + " " + s2);
  _candidate = c;
  _initialCandidate = true;
  _initialSynonym1Candidate = true;
  _initialSynonym2Candidate = true;
  _synonym1 = s1;
  _synonym2 = s2;
  }

 public void setCandidate (OntologyInstance c, double score)
 {
  _candidate = c;
  _score = score;
  }

 public OntologyInstance getOrganCandidate () { return _candidate; }

 OntologyInstance getHomologyGroup () { return _homologyGroup; }

 public void addToReceivedScore (double value, String key)
 {
  Double d = _receivedScoreContribution.get (key);
  if (d == null) _receivedScoreContribution.put (key, value);
  else _receivedScoreContribution.put (key, d + value);
  if (_receivedScore == - 1) _receivedScore = 0;
  _receivedScore += value;
  //System.out.println ("_receivedScore new value: " + _receivedScore + " (added for key: " + key + " : " + value + ")");
  }

 public void cleanContributorToReceivedScore (String key)
 {
  Double d = _receivedScoreContribution.get (key);
  if (d != null)
     {
      _receivedScore -= d;
      _receivedScoreContribution.remove (key);
      }
  }

 public double getIntrinsicScore ()
 {
  if (_score == -1)
     {
      if ((_synonym1 != null) && (_synonym2 != null))
         return _scorer.score (_synonym1, _synonym2);
      else
         if (_synonym1 != null)
            return _scorer.score (_synonym1, _candidate.getName ());
         else
            if (_synonym2 != null)
               return _scorer.score (_instance.getName (), _synonym2);
      return _scorer.score (_instance.getName (), _candidate.getName ());
      }
  else return _score;
  }

 public boolean isValidated () { return _isValidated; }

 public void validate ()
 {
  _initialCandidate = false;
  _initialSynonym1Candidate = false;
  _initialSynonym2Candidate = false;
  _isValidated = true;
  _score = ((PairingScorer) _scorer).getValidatedHomologyScore ();
  _receivedScore = - 1;
  }

 public void setGroup (Vector <PairingScore> group) { _group = group; }

 /**
  * @return The full score of the alignment of the two objects.
  */
 public double getScore ()
 {
  //System.err.println ("Scoring " + _candidate.getName () + " (" + _candidate.getId () + ") and " + _instance.getName () + " (" + _instance.getId ()+ ")" + " " + _initialSynonym1Candidate + " " + _initialSynonym2Candidate);

  if (_initialSynonym1Candidate && _initialSynonym2Candidate)
     {
      //System.err.println ("Score (initial synonym12 candidate) is: " + (_scorer.score (_synonym1, _synonym2) + (_receivedScore == - 1 ? 0 : _receivedScore / 2)));
      return (_scorer.score (_synonym1, _synonym2) + (_receivedScore == - 1 ? 0 : _receivedScore / 2));
      }
  else
     if (_initialSynonym1Candidate)
        {
         //System.out.println ("Score (initial synonym1 candidate) is: " + (_scorer.score (_synonym1, _candidate.getName ()) + (_receivedScore == - 1 ? 0 : _receivedScore / 2)));
         return (_scorer.score (_synonym1, _candidate.getName ()) + (_receivedScore == - 1 ? 0 : _receivedScore / 2));
         }
     else
        if (_initialSynonym2Candidate)
           {
            //System.out.println ("Score (initial synonym2 candidate) is: " + (_scorer.score (_instance.getName (), _synonym2) + (_receivedScore == - 1 ? 0 : _receivedScore / 2)));
            return (_scorer.score (_instance.getName (), _synonym2) + (_receivedScore == - 1 ? 0 : _receivedScore / 2));
            }


  // Ugly turnaround to test a different starting score for homonyms:
  if (_initialCandidate)
     {
      //System.out.println ("Score (initial candidate) is: " +
          //(_scorer.score (_instance.getName (), _candidate.getName ()) + (_receivedScore == - 1 ? 0 : _receivedScore / 2)));
      //System.out.println ("\tIntrinsic score: " + _scorer.score (_instance.getName (), _candidate.getName ()));

      return (_scorer.score (_instance.getName (), _candidate.getName ()) + (_receivedScore == - 1 ? 0 : _receivedScore / 2));
      }

  if (_score != -1)
     {
    //  System.out.println ("\tHomology");
      // Homology. No received score to give.
   //System.out.println ("Score (homology only?) is: " + _score);
      return _score;
      }
  else
     if ((_group == null) && (_receivedScore == - 1))
        {
      //   System.out.println ("\tNo received");
         // No received score:
         //System.out.println ("Score is (nothing received): " + _scorer.score (_instance.getName (), _candidate.getName ()));
         return _scorer.score (_instance.getName (), _candidate.getName ());
         }
     else
        {
         //System.out.println ("\tReceived");
//         if (_group == null) // special case, propagation after initialization (no group created)
//            {
//             System.out.println ("Score is (should not be reach???): " + (_scorer.score (_instance, _candidate) + (_receivedScore == - 1 ? 0 : _receivedScore / 2)));
//             return _scorer.score (_instance, _candidate) + (_receivedScore == - 1 ? 0 : _receivedScore / 2);
//             }
         //System.out.println ("Score is: " + (_scorer.score (_instance.getName (), _candidate.getName ()) + (_receivedScore == - 1 ? 0 : _receivedScore / ((_group.size () + 1) * 2))));
         return _scorer.score (_instance.getName (), _candidate.getName ()) + (_receivedScore == - 1 ? 0 : _receivedScore / ((_group.size () + 1) * 2));
         }
  }

 /**
  *
  */
 public OntologyInstance getInstance () { return _instance; }

 /**
  * 
  */
 public static void setScorer (PairingScorer ps) { _scorer = ps; }

 public void invalidateSelf ()
 {
 if (_group != null)
  for (int i = 0; i < _group.size (); i++)
      if (_group.get (i) != this)
         _group.get (i).remove (this);
  }

 private void remove (PairingScore score)
 {
  _group.remove (score);
  }
 } // class Score