package visualization;

import java.awt.Dimension;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JTabbedPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * This class is used to allow display of a component in two different JTabbedPane
 * objects.
 * Due to a java limitation, a JComponent can have only one ancestor, it is not possible
 * simply add a component to two JTabbedPane to visualize it in both panels.
 * Moreover, I consider that we never want to visualize the same object simultaneously in
 * both panels. The only to do so is a duplication of the component, in the general case
 * it is not a reasonable approach.
 *
 * The idea is the following: when the user select a tab, we intercept the event and we
 * add to the tab its actual content. Once another tab is selected, the content is deleted
 * from the tab, and thus it becomes available for display in the other panel.
 * A problem arises when the selected tab contains a component that is already displayed
 * in the other panel. I tackle this by changing the display of the other panel, turning
 * it into the description associated to the component selected. This is the policy for
 * the case I designed this component for. If there is not another object associated to
 * the selected component, another component should be randomly selected.
 *
 * @author Gilles Parmentier
 */

public class DualTabbedPane extends JTabbedPane
{
 /**
  * 
  */
 private static final long serialVersionUID = 1L;
 JTabbedPane leftPane_, rightPane_;
 Vector<JComponent> components_; // Vectors of JComponent.
 Vector <String> labels_;
 int left_index_, right_index_;
 
 public DualTabbedPane ()
 {
  leftPane_ = new JTabbedPane ();
  leftPane_.setPreferredSize (new Dimension (400, 600));
  leftPane_.addChangeListener (new DualChangeListener ());
  rightPane_ = new JTabbedPane ();
  rightPane_.setPreferredSize (new Dimension (400, 600));
  rightPane_.addChangeListener (new DualChangeListener ());
  components_ = new Vector<JComponent> ();
  left_index_ = right_index_ = - 1;
  }

 public JTabbedPane getLeftPanel () { return leftPane_; }
 public int getLeftSelectedIndex () { return left_index_; }
 public JTabbedPane getRightPanel () { return rightPane_; }
 public int getRightSelectedIndex () { return right_index_; }
 boolean init = false;

 /**
  * Function to add components in the panels. In this case, jc1 and jc2 are associated,
  * meaning that if we select jc1 on the left side and that it was already displayed on
  * the right side, the display will be jc2 on the right and jc1 on the left.
  */

 public void add (JComponent jc1, String name1, JComponent jc2, String name2)
 { 
  init = true;
  components_.add (jc1);
  components_.add (jc2);

  if (left_index_ != - 1)
     leftPane_.setComponentAt (left_index_, new JLabel ());

  leftPane_.addTab (name1, jc1);
  leftPane_.addTab (name2, new JLabel ());
  leftPane_.setSelectedIndex (leftPane_.getTabCount () - 2);
  left_index_ = leftPane_.getSelectedIndex ();

  if (right_index_ != - 1)
     rightPane_.setComponentAt (right_index_, new JLabel ());

  rightPane_.addTab (name1, new JLabel ());
  rightPane_.addTab (name2, jc2);
  rightPane_.setSelectedIndex (rightPane_.getTabCount () - 1);
  right_index_ = rightPane_.getSelectedIndex ();
  init = false;
  }

 class DualChangeListener implements ChangeListener
 {
  public void stateChanged (ChangeEvent e)
  {
   if (init) return;
   JTabbedPane jtp = (JTabbedPane) e.getSource ();

   if ((leftPane_.getTabCount () == 0) || (rightPane_.getTabCount () == 0)) return;

   if (jtp == leftPane_)
      {
       // Removes the component from the old tab:
       if (left_index_ != - 1)
          {
           leftPane_.setComponentAt (left_index_, new JLabel ());
           }
       // Update the index:
       left_index_ = leftPane_.getSelectedIndex ();
       if (left_index_ == right_index_)
          {
           // C'est la cas chiant, il faut afficher autre chose dans le panel droit...
           // Je pars du principe que les objets sont ajoutes par paires, et du coup
           // j'affiche son compagnon dans l'autre panel. Donc si left_index_ est pair
           // j'y affiche left_index_ + 1, sinon left_index_ - 1.
           if (right_index_ != - 1)
              rightPane_.setComponentAt (right_index_, new JLabel ());
           if (left_index_ % 2 == 0)
              right_index_ = left_index_ + 1;
           else
              right_index_ = left_index_ - 1;
           rightPane_.setSelectedIndex (right_index_);
           rightPane_.setComponentAt (right_index_,
                                             components_.get (right_index_));
           }
       // Display the component in the selected tab:
       leftPane_.setComponentAt (left_index_, components_.get (left_index_));
       }
   else
      {
       // Removes the component from the old tab:
       if (right_index_ != - 1)
          rightPane_.setComponentAt (right_index_, new JLabel ());
       // Update the index:
       right_index_ = rightPane_.getSelectedIndex ();
       if (right_index_ == left_index_)
          {
           // C'est la cas chiant, il faut afficher autre chose dans le panel droit...
           // Je pars du principe que les objets sont ajoutes par paires, et du coup
           // j'affiche son compagnon dans l'autre panel. Donc si left_index_ est pair
           // j'y affiche left_index_ + 1, sinon left_index_ - 1.
           if (left_index_ != - 1)
              leftPane_.setComponentAt (left_index_, new JLabel ());
           if (right_index_ % 2 == 0)
              left_index_ = right_index_ + 1;
           else
              left_index_ = right_index_ - 1;
           leftPane_.setSelectedIndex (left_index_);
           leftPane_.setComponentAt (left_index_,
                                              components_.get (left_index_));
           }
       // Display the component in the selected tab:
       rightPane_.setComponentAt (right_index_, components_.get (right_index_));
       }
   }
  }
 }
