package visualization;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;

import javax.swing.*;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeNode;

import datalayer.OboHomologyReader;
import datalayer.WorkspaceReader;
import datalayer.WorkspaceWriter;

import alignment.HomologyMaker;

import ontology.Annotations;
import ontology.HomologyOntology;
import ontology.OboOntology;
import ontology.Ontology;
import ontology.OntologyClass;
import ontology.OntologyFactory;
import ontology.OntologyInstance;
import ontology.OntologyObject;
import ontology.OntologyProperty;
import ontology.SpeciesOntology;

/**
 * @author Gilles Parmentier
 */

public class Main implements ActionListener, TreeSelectionListener
{
 JFrame _mainFrame;
 JMenuBar _jmb;
 JMenu onto1, onto2;
 boolean _oRefSet, _oWorkSet;
 JMenuItem jmiAlign;
 JRadioButtonMenuItem [] _jrbmiScoring;
 DisplayPropositions _dp;
 JMenuItem _jmiOpen, _jmiFindId, _jmiFindOrphan;
 HomologyMaker _aligner;
 HomologyOntology _ho = null;
 boolean _hasNewHomology = false;

 OntologyInstance _lastReferenceOntologySelectedInstance;
 OntologyInstance _lastWorkingOntologySelectedInstance;

 // Tree rendering:
 private DualTabbedPane _dual_pane;
 private JSplitPane _split_pane;
 private Vector<OntologyTreeView> _trees;
 private Vector<JEditorPane> _htmlPanes;
 private OntologyCellRenderer _ocr;
 private boolean _hasData;

 private LinkOntologyObjectsAction _looa;

 // File opening: :
 private OpenSpeciesOntologyFileAction _ofa;
 //private OpenHomologyOntologyFileAction _hfa;
 private OpenWorkspaceAction _owa;
 //private SaveHomologyOntologyFileAction _sfa;

 // Fields related to the alignment itself:
 private AlignmentFrame _af;

 static final int SIMPLE_MATCHING = 0;
 static final int SIMPLE_WITH_HIERARCHY = 1;
 static final int PARTIAL_MATCHING = 2;

 FindWindow _fw;
 FindOrphanWindow _fow;
 
 JFileChooser _jfc;

 // Workspace data:
 String _workspace_filename = null;
 String _discards_filename = null;
 String _homology_ontology_filename = null;
 String _annotations_filename = null;
 Vector <String> _species_ontology_filename = null;

 boolean _merging_mode = false;

/**
 *
 */
 Main ()
 {
  _ofa = new OpenSpeciesOntologyFileAction ();
  //_hfa = new OpenHomologyOntologyFileAction ();
  //_sfa = new SaveHomologyOntologyFileAction ();
  _looa = new LinkOntologyObjectsAction ();
  _owa = new OpenWorkspaceAction ();
	  
  _mainFrame = new JFrame ("OntoAligner");
  _mainFrame.setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);
  _dp = new DisplayPropositions (_mainFrame);
  makeMenu ();
  _oRefSet = _oWorkSet = false;
  _dual_pane = new DualTabbedPane ();
  JFrame.setDefaultLookAndFeelDecorated (true);

  //Create and set up the window.
  _split_pane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
  _split_pane.setLeftComponent(_dual_pane.getLeftPanel());
  _split_pane.setRightComponent(_dual_pane.getRightPanel());
  _mainFrame.getContentPane().add(_split_pane, BorderLayout.CENTER);

  JToolBar jtb = new JToolBar ();
  makeJTBButtons (jtb);
  _mainFrame.getContentPane ().add (jtb, BorderLayout.PAGE_START);

  // Initializations:
  _trees = new Vector<OntologyTreeView> ();
  _htmlPanes = new Vector<JEditorPane> ();

  _ocr = new OntologyCellRenderer ();

  _hasData = false;
  _fw = new FindWindow (_mainFrame, this);
  _fow = new FindOrphanWindow (_mainFrame, this);
  _jfc = new JFileChooser ();

  _aligner = new HomologyMaker (this);

  //Display the window.
  _mainFrame.pack ();
  _mainFrame.setVisible (true);
  }

 /**
  * @param e
  */
 public void actionPerformed (ActionEvent e)
 {
  JMenuItem source = (JMenuItem)(e.getSource());
  //System.err.println ("Action event detected." + "\n" + "    Event source: " +
  //   source.getText() + " (an instance of " + getClassName(source) + ")");

  if (source.getText ().equals ("Align selected ontologies")) { align (); }
  else
     if (source.getText ().equals ("Find id...")) { find (); }
     else
        if (source.getText ().equals("Save and Exit")) { exit (); }
        else
            if (source.getText ().equals("Find Organ without Homolog...")) { findOrphan (); }
            else
             if (source.getText ().equals ("Merge ontologies...")) { mergeHomologyOntologies (); }
    }

 /**
  * Reads 2 existing homology ontology and merges the homology group.
  *
  */
 private void mergeHomologyOntologies ()
 {
  // Read the two ontologies:
  JFileChooser jfc = new JFileChooser ("Choose first project...");
  jfc.setFileSelectionMode (JFileChooser.DIRECTORIES_ONLY);
  int ret = jfc.showOpenDialog (this._mainFrame);

  if (ret == JFileChooser.APPROVE_OPTION)
     {
      _merging_mode = true;
      File file = jfc.getSelectedFile ();
      String wf1 = file.getAbsolutePath ();
      WorkspaceReader wr = new WorkspaceReader (wf1);
      // Open the homology ontology:
      String hof1 = wr.getHomologyOntologyFilename ();
      HomologyOntology ho1 = null;
      try
          {
           OboHomologyReader ohr = new OboHomologyReader (wf1 + File.separator + hof1);
           ho1 = ohr.getOntology ();
           createHomologyView  (ho1);
           }
      catch (Exception e) { e.printStackTrace (); }
      // Second project:
      jfc = new JFileChooser ("Choose 2nd project...");
      jfc.setFileSelectionMode (JFileChooser.DIRECTORIES_ONLY);
      ret = jfc.showOpenDialog (this._mainFrame);
      if (ret == JFileChooser.APPROVE_OPTION)
         {
          File file2 = jfc.getSelectedFile ();
          String wf2 = file2.getAbsolutePath ();
          WorkspaceReader wr2 = new WorkspaceReader (wf2);
          // Open the homology ontology:
          String hof2 = wr2.getHomologyOntologyFilename ();
          HomologyOntology ho2 = null;
          try
              {
               OboHomologyReader ohr = new OboHomologyReader (wf2 + File.separator + hof2);
               ho2 = ohr.getOntology ();
               createHomologyView  (ho2);
               }
          catch (Exception e) { e.printStackTrace (); }

          // Now we can merge the ontologies:
          HomologyOntology result = new HomologyOntology (ho1, ho2);

          // Save the result:
          jfc.setFileSelectionMode (JFileChooser.DIRECTORIES_ONLY);
          ret = jfc.showOpenDialog (this._mainFrame);
          if (ret == JFileChooser.APPROVE_OPTION)
             {
              File file3 = jfc.getSelectedFile ();
              String wf3 = file3.getAbsolutePath ();
              WorkspaceWriter wr3 = new WorkspaceWriter (wf3, wr, wr2, result);
              }
          }
      }
  else
     {
      return;
      }
  }

 /**
  * This function constructs the tree view of the ontology.
  * It's here because I don't know yet where to put it. I probably miss an intermediate
  * class in my model, between OntologyTreeView and the ontology itself.
  */
 private DefaultTreeModel makeTree (Ontology _ontology, DefaultMutableTreeNode _root, String property, String inv_property)
 {
  System.out.println ("Creating tree view of an ontology...");

  // Add the classes to the root:
  Vector <OntologyClass> classes = _ontology.getClasses ();

  for (int i = 0; i < classes.size (); i++)
      {
       OntologyClass oc = (OntologyClass) classes.get (i);
       OntologyClassView ocv = new OntologyClassView (oc);
       DefaultMutableTreeNode son = new DefaultMutableTreeNode (ocv);
       _root.add (son);
       }

  // Add the properties:
  Vector <OntologyProperty> properties = _ontology.getProperties ();

  for (int i = 0; i < properties.size (); i++)
      {
       OntologyProperty op = (OntologyProperty) properties.get (i);
       OntologyPropertyView opv = new OntologyPropertyView (op);
       DefaultMutableTreeNode son = new DefaultMutableTreeNode (opv);
       _root.add (son);
       }

  // Finally, add the instances:
  java.util.List <OntologyInstance> instances = _ontology.getInstances ();
  System.err.println ("Main; makeTree: " + instances.size () + " instances");
  // Find all instances without parent for the chosen property;
  for (int i = 0; i < instances.size (); i++)
      {
       //System.err.println ("i: " + i);
       OntologyInstance oi = (OntologyInstance) instances.get (i);
       Vector <String> parents = oi.getObjectOfPropertyByPropertyIdAsVector (property);
       if (parents == null)
          {
	   //System.err.println (oi.getId () + " is a root");
           // This node is a root for the part_of property.
           OntologyInstanceView oiv = new OntologyInstanceView (oi);
           DefaultMutableTreeNode son = new DefaultMutableTreeNode (oiv);
           _root.add (son);
           //Recursively add its descendant:
           addInstancesToTree (_ontology, oi, son, property, inv_property, "");
           }
       else
          {
           //System.err.println ("Not a root: " + oi.getId () + "(" + (String) parents.elementAt(0)+ ")");
           }
      }
  DefaultTreeModel tm = new DefaultTreeModel (_root);
  return tm;
  }

 /**
  * Recursively adds the instances in the <code>DefaultMutableTreeNode</code> hierarchy.
  * This function explicitly uses the part_of property as the hierarchy criteria.
  * As aboce, this function should probably be somewhere else.
  */
 private void addInstancesToTree (Ontology _ontology, OntologyInstance oiParent, DefaultMutableTreeNode dmtnParent, String property, String inv_property, String dec)
 {
  //System.err.println (dec + "addInstancesToTree_partOf: " + oiParent.getName () + "; " + oiParent.getId ());
  //dec += "\t";
  //oiParent.printProperties ();
  // List all the sons of parent for the part_of property:
  Vector <String> sons = oiParent.getObjectOfPropertyByPropertyIdAsVector (inv_property);
  if (sons != null)
     for (int i = 0; i < sons.size (); i++)
         {
          OntologyInstance oi = _ontology.getInstanceById ((String) sons.get (i));
          // oi may be null as not all instances are obliged to be members of the current
          // ontology.
          if (oi != null)
             {
              //System.err.println ("Adding a son to " + oiParent.getName () + ": " + oi.getName ());
              OntologyInstanceView oiv = new OntologyInstanceView (oi);
              DefaultMutableTreeNode son = new DefaultMutableTreeNode (oiv);
              dmtnParent.add (son);
              addInstancesToTree (_ontology, oi, son, property, inv_property, dec);
              }
          }
  }

 /**
  * @param jtb
  */
 private void makeJTBButtons (JToolBar jtb)
 {
  JButton jb = new JButton ();
  jb.setAction (_ofa);
  jb.setText ("");
  jb.setToolTipText ("Open an obo ontology file");
  jtb.add (jb);
  jb = new JButton ();
  jb.setAction (_looa);
  jb.setText ("");
  jb.setToolTipText ("Will propose some link creation if you selected an object in both ontologies");
  jtb.add(jb);
  }

 /**
  * 
  */
 void makeMenu ()
 {
  _jmb = new JMenuBar ();

  // "File" menu:
  JMenu menu = new JMenu ("File");
  _jmb.add (menu);


  // "Open" part:
  _jmiOpen = new JMenuItem (_ofa);
  menu.add (_jmiOpen);

  JMenuItem jmi = new JMenuItem (_owa);
  menu.add (jmi);

//  _jmiHomologyOpen = new JMenuItem (_hfa);
//  _jmiHomologyOpen.setEnabled (false);
//  menu.add (_jmiHomologyOpen);
  
  menu.addSeparator ();

//  // "Save" part:
//  _jmiHomologySave = new JMenuItem (_sfa);
//  _jmiHomologySave.setEnabled (false);
//  menu.add (_jmiHomologySave);
//
//  menu.addSeparator ();
  
  // "Exit" part:
  jmi = new JMenuItem ("Save and Exit");
  jmi.addActionListener (this);
  menu.add (jmi);

  menu = new JMenu ("Find");
  _jmb.add (menu);

  _jmiFindId = new JMenuItem ("Find id...");
  _jmiFindId.addActionListener (this);
  menu.add (_jmiFindId);

  _jmiFindOrphan = new JMenuItem ("Find Organ without Homolog...");
  _jmiFindOrphan.addActionListener (this);
  menu.add (_jmiFindOrphan);

  menu = new JMenu ("Alignment");
  _jmb.add (menu);

  JMenu score = new JMenu ("Scoring method");
  _jrbmiScoring = new JRadioButtonMenuItem [3];
  ButtonGroup bg = new ButtonGroup ();
  _jrbmiScoring [0] = new JRadioButtonMenuItem ("Perfect matching");
  //_jrbmiScoring [0].addActionListener (this);
  bg.add (_jrbmiScoring [0]);
  score.add (_jrbmiScoring [0]);
  _jrbmiScoring [1] = new JRadioButtonMenuItem ("Perfect matching with hierarchy modificator");
  //_jrbmiScoring [1].addActionListener (this);
  bg.add (_jrbmiScoring [1]);
  score.add (_jrbmiScoring [1]);
  _jrbmiScoring [2] = new JRadioButtonMenuItem ("Partial matching");
  //_jrbmiScoring [2].addActionListener (this);
  bg.add (_jrbmiScoring [2]);
  score.add (_jrbmiScoring [2]);
  _jrbmiScoring [2].setSelected (true);
  menu.add (score);

  jmiAlign = new JMenuItem ("Align selected ontologies");
  jmiAlign.addActionListener (this);
  jmiAlign.setEnabled (false);
  menu.add (jmiAlign);

  menu = new JMenu ("Homology");
  jmi = new JMenuItem ("Merge ontologies...");
  jmi.addActionListener (this);
  menu.add (jmi);
  _jmb.add (menu);

  _mainFrame.setJMenuBar (_jmb);
  }

 protected String getClassName (Object o)
 {
  String classString = o.getClass ().getName ();
  int dotIndex = classString.lastIndexOf (".");
  return classString.substring (dotIndex + 1);
  }

 // Create and display GUI:
 private static void createAndShowGUI ()
 {
  try
      {
       UIManager.setLookAndFeel (UIManager.getCrossPlatformLookAndFeelClassName ());
       }
  catch (Exception e) { e.printStackTrace(); }
  Main m = new Main ();
  }

 private void addTreeToUI (OntologyTreeView otv, String label)
 {
  _trees.add (otv);
  JEditorPane htmlPane = new JEditorPane ();
  htmlPane.setEditorKit (new HTMLEditorKit ());
  htmlPane.setEditable (false);
  _htmlPanes.add (htmlPane);
  Dimension ms = new Dimension (200, 300);
  JScrollPane htmlView = new JScrollPane (htmlPane);
  htmlView.setMinimumSize (ms);
  htmlView.setPreferredSize (new Dimension (300, 600));
  JScrollPane treeView = new JScrollPane (otv);
  treeView.setMinimumSize (ms);
  treeView.setPreferredSize (new Dimension (300, 600));
  _dual_pane.add (treeView, label, htmlView, "Detailed term " + label);
  }

 private void initUIWithTree (OntologyTreeView otv, String label)
 {
  _trees.add (otv);
  JEditorPane htmlPane = new JEditorPane ();
  htmlPane.setEditorKit (new HTMLEditorKit ());
  htmlPane.setEditable (false);
  _htmlPanes.add (htmlPane);
  Dimension ms = new Dimension (200, 300);
  JScrollPane htmlView = new JScrollPane (htmlPane);
  htmlView.setMinimumSize (ms);
  htmlView.setPreferredSize (new Dimension (300, 600));
  JScrollPane treeView = new JScrollPane (otv);
  treeView.setMinimumSize (ms);
  treeView.setPreferredSize (new Dimension (300, 600));
  _dual_pane.add (treeView, label, htmlView, "Detailed term " + label);
  _hasData = true;
  }

 /**
  * Function belonging to the <code>TreeSelectionListener</code> interface. It is called
  * when a node is selected. It retrieves the good tree and displays the information for
  * the selected node in the good panel.
  */
 public void valueChanged (TreeSelectionEvent e)
 {
  // Ici il faudrait que je recupere l'arbre sur lequel l'utilisateur a clique. A partir
  // de la, je retrouve le htmlPane dans lequel je dois afficher.
  JTree the_tree = (JTree) e.getSource ();
  int index = this.getTreeIndex (the_tree);
  DefaultMutableTreeNode node = (DefaultMutableTreeNode)
                                                 the_tree.getLastSelectedPathComponent ();
  if (node == null) return;
  try
     {
      // Affiche les details concernant le noeud selectionne :
      OntologyObjectView oov = (OntologyObjectView) node.getUserObject ();
      // C'est la que je dois selectionner le bon htmlPane :
      _htmlPanes.get (index).setText (oov.display ());

      // Je stocke des informations sur l'ontologie et l'objet qui ont ete selectionne,
      // pour pouvoir m'en servir si l'utilisateur veut creer un lien d'homologie ad hoc :

      OntologyObject oo = oov.getObject();
      // Verifier si cet objet est un OntologyInstance :
 if (oo.getClass ().getName ().equals ("ontology.OntologyInstance"))
 {
// C'est bien parti, j'ai un OntologyInstance. Maintenant je dois determiner a quelle ontologie il appartient.
SpeciesOntology ref = (SpeciesOntology) _aligner.getReferenceOntology();
if (ref != null)
{
if (ref.getInstanceById (oo.getId ()) != null)
_lastReferenceOntologySelectedInstance = (OntologyInstance) oo;
else
{
SpeciesOntology work = (SpeciesOntology) _aligner.getWorkingOntology();
if (work.getInstanceById (oo.getId ()) != null)
_lastWorkingOntologySelectedInstance = (OntologyInstance) oo;
}
 }
 }
      }
  catch (ClassCastException cce) {}
  }

 /**
  * Returns the index of the given tree in the internal trees <code>Vector</code>.
  */
 private int getTreeIndex (JTree tree)
 {
  for (int i = 0; i < _trees.size (); i++)
      if (_trees.get (i).hasTree (tree))
         return i;
  return - 1;
  }

 public void openSpeciesOntologyFile (File file)
 {
      //System.out.println ("Opening: " + file.getName ());
      OntologyFactory of = new OntologyFactory ();
      if (_species_ontology_filename == null)
         _species_ontology_filename = new Vector <String> ();
      _species_ontology_filename.add (file.getName ());
      //long start = System.currentTimeMillis ();
      
      Ontology o = of.getOboOntology(file.getAbsolutePath ());

      DefaultMutableTreeNode r = new DefaultMutableTreeNode (null);
      DefaultTreeModel tm = makeTree (o, r, ((OboOntology) o).getPartOfPropertyId (), "inv_" + ((OboOntology) o).getPartOfPropertyId ());
      final OntologyTreeView otv = new OntologyTreeView (Main.this, _ocr, tm, o, r);
      if (_hasData)
         {
          addTreeToUI (otv, file.getName ());
          }
      else
         {
          initUIWithTree (otv, file.getName ());
          }
      if (_oRefSet)
         {
          _aligner.setWorkingOntology ((SpeciesOntology) o);
          _oWorkSet = true;
          _ho = new HomologyOntology (null);
          _ho.addSpeciesOntology ((SpeciesOntology) _aligner.getWorkingOntology ());
          _ho.addSpeciesOntology ((SpeciesOntology) _aligner.getReferenceOntology ());
          _aligner.setHomologyOntology (_ho);
          //_jmiHomologyOpen.setEnabled (true);
          jmiAlign.setEnabled (true);
          }
      else
         {
          _aligner.setReferenceOntology((SpeciesOntology) o);
          _oRefSet = true;
          }
      //long end = System.currentTimeMillis ();
      //long time = (end - start) / 1000;

      //System.err.println ("total opening time: " + time);



  }

 /**
  * Opens an existing workspace.
  *
  */
 public void openWorkspace ()
 {
  JFileChooser jfc = new JFileChooser ();
  jfc.setFileSelectionMode (JFileChooser.DIRECTORIES_ONLY);

  int ret = jfc.showOpenDialog (this._mainFrame);
  if (ret == JFileChooser.APPROVE_OPTION)
     {
      File file = jfc.getSelectedFile ();
      _workspace_filename = file.getAbsolutePath ();
      WorkspaceReader wr = new WorkspaceReader (_workspace_filename);

      // Species ontology:
      _species_ontology_filename = wr.getSpeciesFilename ();
      for (int i = 0; i < _species_ontology_filename.size (); i++)
          {
           System.err.println ("Opening: " + _species_ontology_filename.get (i));
           OntologyFactory of = new OntologyFactory ();

           long start = System.currentTimeMillis ();

           Ontology o = of.getOboOntology (_species_ontology_filename.get (i));

           DefaultMutableTreeNode r = new DefaultMutableTreeNode (null);
           DefaultTreeModel tm = makeTree (o, r, ((OboOntology) o).getPartOfPropertyId (), "inv_" + ((OboOntology) o).getPartOfPropertyId ());
           OntologyTreeView otv = new OntologyTreeView (this, _ocr, tm, o, r);
           if (_hasData) { addTreeToUI (otv, new File (_species_ontology_filename.get (i)).getName ()); }
           else { initUIWithTree (otv, new File (_species_ontology_filename.get (i)).getName ()); }
           if (_oRefSet)
              {
               System.err.println ("Setting woking onto");
               _aligner.setWorkingOntology ((SpeciesOntology) o);
               _oWorkSet = true;
               _ho = new HomologyOntology (null);
               _ho.addSpeciesOntology ((SpeciesOntology) _aligner.getWorkingOntology ());
               _ho.addSpeciesOntology ((SpeciesOntology) _aligner.getReferenceOntology ());
               _aligner.setHomologyOntology (_ho);
               jmiAlign.setEnabled (true);
               }
           else
              {
            System.err.println ("Setting reference onto");
               _aligner.setReferenceOntology ((SpeciesOntology) o);
               _oRefSet = true;
               }

           long end = System.currentTimeMillis ();
           long time = (end - start) / 1000;
           System.err.println ("total opening time: " + time);
           }
      
      _discards_filename = _workspace_filename + File.separator +"discards";
      _annotations_filename = _workspace_filename + File.separator + "annot";
      this.readAnnotations (_annotations_filename);

      // Restore data;
      int scheme_value = -1;
      for (int j = 0; j < _jrbmiScoring.length; j++)
          if (_jrbmiScoring [j].isSelected ())
             scheme_value = j;
      _af = new AlignmentFrame (_aligner, this);
      _af.setScoringScheme (scheme_value);
      _af.restoreAlignment (_workspace_filename + File.separator + "decisions");
      }
  }

 private void createHomologyView (HomologyOntology ho)
 {
  DefaultMutableTreeNode r = new DefaultMutableTreeNode (null);
  DefaultTreeModel tm = makeTree (ho, r, ((OboOntology) ho).getPartOfPropertyId (), "inv_" + ((OboOntology) ho).getPartOfPropertyId ());
  OntologyTreeView otv = new OntologyTreeView (this, _ocr, tm, ho, r);
  addTreeToUI (otv, "Homology ontology");
  }

public void setHasNewHomology ()
 {
  _hasNewHomology = true;
  //_jmiHomologySave.setEnabled (true);
  }

 public Vector <TreeNode> searchOrphanInOtontologies ()
 {
  Vector <TreeNode> res = new Vector <TreeNode> ();
  for (int i = 0; i < _species_ontology_filename.size (); i++)
      {
       // D'abord je dois verifier si une instance est selectionnee. Si c'est le cas,
       // on cherche uniquement sous cette instance.
       // Si aucune instance n'est selectionnee, on cherche a partir de la racine.
       }
  return res;
  }

 public Vector<TreeNode> searchInOntologies (String text, String ontology)
 {
  Vector<TreeNode> res = new Vector<TreeNode> ();
  System.err.println ("Searching in " + ontology);
  if (ontology.equals ("both"))
     {
      }
  else
  {
   // Selectionner la bonne ontologie :
   for (int i = 0; i < _species_ontology_filename.size (); i++)
       {
        System.out.println ("Checking: " + _species_ontology_filename.get (i));
        if (ontology.equals (_species_ontology_filename.get (i)))
           {
            //OntologyTreeView otv = _trees.get (2 * i);
            OntologyTreeView otv = _trees.get (i);
            res = otv.find (text);
            if (res.size () > 1)
               {
                // Activer le bouton next.
                _fw.setTitle (res.size () + " results");
                _fw.activateNext ();
                }
            else
               if (res.size () == 0) { System.err.println ("Nothing found"); _fw.setTitle ("No result!"); }
               else
                _fw.setTitle ("1 result");
            break;
            }
        }
  }
  return res;
  }

 public Vector <TreeNode> searchOrphanInOntologies (String choice)
 {
  Vector <TreeNode> res = new Vector <TreeNode> ();
  for (int i = 0; i < _species_ontology_filename.size (); i++)
      {
       if (choice.equals(_species_ontology_filename.get (i)))
          {
	   OntologyTreeView otv = _trees.get (i);
	   res = otv.findOrphan ();
	   if (res.size () > 1)
	       _fow.activateNext ();
           }
       }
  return res;
  }

 public void iterateSearchOrphanInOntologies(String string) {
     // TODO Auto-generated method stub
     
 }

 public void iterateSearchInOntologies (String ontology)
 {
  for (int i = 0; i < _species_ontology_filename.size (); i++)
      {
       System.out.println ("Checking: " + _species_ontology_filename.get (i));
       if (ontology.equals (_species_ontology_filename.get (i)))
          {
           //OntologyTreeView otv = _trees.get (2 * i);
           OntologyTreeView otv = _trees.get (i);
           otv.iterateFind ();
           break;
           }
       }
  }

 public void linkObjects ()
 {
  // First we check that an element is selected in both ontologies
  if (_lastReferenceOntologySelectedInstance != null)
     {
      if (_lastWorkingOntologySelectedInstance != null)
         {
          System.err.println ("Valid link creation requested !");
          System.err.println ("" + _lastReferenceOntologySelectedInstance.getId ());
          System.err.println ("" + _lastWorkingOntologySelectedInstance.getId ());

          // Ouvrir une fenetre de dialogue sur la nature du lien a creer :
          Object [] choices = {"Homology link", "Loose homology link", "Invalidate this pair"};
          String s = (String)JOptionPane.showInputDialog (_mainFrame,
          "What relation do you want to add between\n" +
          _lastReferenceOntologySelectedInstance.getId () +
          " and " + _lastWorkingOntologySelectedInstance.getId () + "?",
          "Customized Dialog", JOptionPane.QUESTION_MESSAGE, null, choices, "ham");
          if (s != null)
             {
              if (s.equals("Homology link"))
                 {
                  _aligner.createHomology (_lastWorkingOntologySelectedInstance.getId (), _lastReferenceOntologySelectedInstance.getId (), false);
                  }
              else
                 if (s.equals ("Invalidate this pair"))
                    {
                     _aligner.createInvalidate (_lastWorkingOntologySelectedInstance.getId (), _lastReferenceOntologySelectedInstance.getId ());
                     }
                 else
                    if (s.equals ("Strict homology link"))
                       {
                        _aligner.createHomology (_lastWorkingOntologySelectedInstance.getId (), _lastReferenceOntologySelectedInstance.getId (), true);
                        }
              }
          else
             {
              // L'utilisateur a clique sur cancel, on ne fait rien.
              }
          }
      else
      { System.err.println ("working object is null"); }
      }
  else
     {
      System.err.println ("reference object is null");
      }
  }

 public static void main (String [] args)
 {
  javax.swing.SwingUtilities.invokeLater (new Runnable () { public void run () { createAndShowGUI (); } });
  }

 /**
  * On exit: check if the user wants to save the current workspace. If yes, check if the
  * workspace was already existing (i.e. the user started its work by opening it). If so,
  * we just overwrite the existing files.
  * If the workspace is a new one, ask the name to the user and create all the
  * corresponding files.
  */
 private void exit ()
 {
  if (_merging_mode) { System.exit (0); }
  // Check if 2 species were opened and if some alignement started.
  int ret = JOptionPane.showConfirmDialog (_mainFrame,
                          "Do you want to save the current workspace?", "Save workspace?",
                                                        JOptionPane.YES_NO_CANCEL_OPTION);
  if (ret == JOptionPane.OK_OPTION)
     {
      if (_workspace_filename != null)
         {
          _aligner.saveHomologyOntology (_workspace_filename + File.separator + "homology.obo");
          try
             {
              _aligner.saveDiscards (_discards_filename);
              }
          catch (IOException e) { e.printStackTrace(); }
          writeAnnotations (_aligner.getAnnotations (), _annotations_filename);
          //_aligner.saveWorkspace (_workspace_filename, _discards_filename, _annotations_filename);
          if (_af.hasDecisions ()) { _af.saveDecisions (_workspace_filename); }
          }
      else
         {
          // Ask the user a name for the workspace.
          _jfc.setSelectedFile (new File ("workspace"));
          int returnVal = _jfc.showSaveDialog (this._mainFrame);

          if (returnVal == JFileChooser.APPROVE_OPTION)
             {
              File file = _jfc.getSelectedFile ();
              // Check if the file has a .xml extension:
              String sfile = file.getAbsolutePath ();
//              String ext = sfile.substring (sfile.length () - 4, sfile.length ());
              String basename = "";
//              if (ext.equals (".xml"))
//                 basename = sfile.substring (0, sfile.length () - 4);
//              else
//                 {
//                  System.err.println ("Extension is: " + ext);
                  basename = sfile;
//                  }
//              System.err.println ("basename is: " + basename);
              _aligner.saveWorkspace (basename, basename + File.separatorChar + "discards", basename + File.separatorChar + "annot");
              _aligner.saveHomologyOntology (basename + File.separator + "homology.obo");
              try { _aligner.saveDiscards (basename + File.separatorChar + "discards"); }
              catch (IOException e) { e.printStackTrace (); }
              writeAnnotations (_aligner.getAnnotations (), basename + File.separatorChar + "annot");
              if ((_af != null) && _af.hasDecisions ()) { _af.saveDecisions (basename); }
              }
          }
      }
  System.exit (0);
  }

 /**
  * 
  */
 private void readAnnotations (String basename)
 {
  System.err.println ("Reading annotations");

  // First, check if we really have annotations:
  File f = new File (basename);
  if (f.exists ())
     {
      try
          {
           BufferedReader br = new BufferedReader (new FileReader (f));
           String str;
           System.err.println ("File exists");
           while ((str = br.readLine ()) != null)
                 {
                  String worker = str;
                  String reference = br.readLine ();
                  String author = br.readLine ();
                  String link = br.readLine ();
                  String comments = br.readLine ();
                  String str_stat = br.readLine ();
                  _aligner.createAnnotation (reference, worker, author, link, comments, new Integer (str_stat).intValue ());
                  System.err.println ("An annotation has been added");
                  }
           }
      catch (Exception e) {}
      }
  }

 /**
  *
  */
 public void writeAnnotations (Vector <Annotations> annotations, String basename)
 {
  if (annotations.size () == 0) { System.err.println ("No annotation written"); return;}
  // Currently I completely overwrite the existing file.

  File f = new File (basename);

  try
     {
      BufferedWriter bw = new BufferedWriter (new FileWriter (f));
      for (int i = 0; i < annotations.size (); i++)
          {
           System.err.println ("Writing an annotation");
           Annotations a = annotations.get(i);
           bw.write (a.getWorker ().getId ());
           bw.newLine ();
           bw.write (a.getReference ().getId ());
           bw.newLine ();
           bw.write (a.getAuthor ());
           bw.newLine ();
           String str = a.getLink ();
           str = str.replace ("\n", " ");
           bw.write (str);
           bw.newLine ();
           str = a.getComments ();
           str = str.replace ("\n", " ");
           bw.write (str);
           bw.newLine ();
           bw.write ("" + a.getStatus ());
           bw.newLine ();
           }
      bw.close ();
      }
  catch (Exception e) {}
  }

 /**
  * Simple find facility.
  */
 private void find ()
 {
  if ((_species_ontology_filename != null) && (_species_ontology_filename.size () > 0))
     {
      _fw.setLocationRelativeTo (_mainFrame);
      _fw.setChoices (_species_ontology_filename);
      _fw.setTitle ("Find name or id...");
      _fw.setVisible (true);
      }
  else
     {
      System.err.println ("Find is not active currently");
      }
  }

 /**
  * 
  */
 private void findOrphan ()
 {
  if ((_species_ontology_filename != null) && (_species_ontology_filename.size () > 0))
     {
      _fow.setLocationRelativeTo (_mainFrame);
      _fow.setChoices (_species_ontology_filename);
      _fow.setVisible (true);
      }
  else
     {
      System.err.println ("Find is not active currently");
      }
  }

 private void align ()
 {
  int scheme_value = -1;
  for (int i = 0; i < _jrbmiScoring.length; i++)
  if (_jrbmiScoring [i].isSelected ())
	 scheme_value = i;

  _af = new AlignmentFrame (_aligner, this);
  _af.setScoringScheme (scheme_value);
  _af.setVisible (true);
  new Thread (new Runnable () { public void run () {
      _af.align ();
  
  // Optional part for testing purposes only:
  // Open the file that contains the manually validated pairs:
  Vector <String []> pairs = new Vector <String []> ();
  try
      {
       BufferedReader bf = new BufferedReader (new FileReader (new File ("data.csv")));
       String str = bf.readLine ();
       while (str != null)
             {
	      // We have to do this as the lines may contain more than 2 terms (comments):
	      String [] pair = new String [2];
	      String [] words = str.split (";");
	      pair [0] = words [0];
	      pair [1] = words [1];
	      pairs.add (pair);
	      str = bf.readLine ();
              }
       }
  catch (Exception e) { e.printStackTrace (); }
  // end of optional part
  
  _af.setValidatedList (pairs);
  //_af.validateFromList();

      _af.displayResults ();
      }}).start ();
  }

 class OpenSpeciesOntologyFileAction extends AbstractAction
 {
  /**
   * 
   */
  private static final long serialVersionUID = 1L;

  public OpenSpeciesOntologyFileAction ()
  { super ("Open Species Ontology", new ImageIcon (Main.class.getResource ("Open24.gif"))); }

  public void actionPerformed (ActionEvent e)
  {

	  int returnVal = _jfc.showOpenDialog (_mainFrame);
	  if (returnVal == JFileChooser.APPROVE_OPTION)
	     {
	      final File file = _jfc.getSelectedFile();
	  final BlockingDialog bf = new BlockingDialog (_mainFrame, "Opening ontology", "Opening ontology: " + file.getName ());
	  SwingUtilities.invokeLater (new Runnable () { public void run () {
      
      bf.setVisible (true);
	   openSpeciesOntologyFile (file); 
 	  bf.setVisible (false);
	  } } );
	     }
   }
  }

 class OpenWorkspaceAction extends AbstractAction
 {
  /**
   * 
   */
  private static final long serialVersionUID = 1L;

  public OpenWorkspaceAction ()
  { super ("Open Workspace"); }

  public void actionPerformed (ActionEvent e) { openWorkspace (); }
  }

 class LinkOntologyObjectsAction extends AbstractAction
 {
  /**
   *
   */
  private static final long serialVersionUID = 1L;

  public LinkOntologyObjectsAction ()
  {
   super ("Link objects", new ImageIcon (Main.class.getResource ("OntologiesLink24.gif")));
   }

  public void actionPerformed (ActionEvent e) { linkObjects (); }
  }
 };